import { QueryCompositeFilterConstraint, Transaction, and, average, collection, collectionGroup, count, deleteDoc, doc, endAt, endBefore, query as firestoreQuery, getAggregateFromServer, getDoc, getDocs, limit, limitToLast, onSnapshot, or, orderBy, setDoc, startAfter, startAt, sum, where, writeBatch, } from '@firebase/firestore';
import { collectionPath, docPath, } from 'firestore-repository/schema';
import { assertNever } from 'firestore-repository/util';
export class Repository {
    collection;
    db;
    constructor(collection, db) {
        this.collection = collection;
        this.db = db;
    }
    async get(id, options) {
        const doc = await (options?.tx ? options.tx.get(this.docRef(id)) : getDoc(this.docRef(id)));
        return this.fromFirestore(doc);
    }
    getOnSnapshot(id, next, error, complete) {
        return onSnapshot(this.docRef(id), {
            next: (doc) => {
                next(this.fromFirestore(doc));
            },
            error: (e) => error?.(e),
            complete: () => {
                complete?.();
            },
        });
    }
    async list(query) {
        const { docs } = await getDocs(toFirestoreQuery(this.db, query));
        return docs.map((doc) => 
        // biome-ignore lint/style/noNonNullAssertion: query result item should not be null
        this.fromFirestore(doc));
    }
    listOnSnapshot(query, next, error, complete) {
        return onSnapshot(toFirestoreQuery(this.db, query), {
            next: ({ docs }) => {
                // biome-ignore lint/style/noNonNullAssertion: query result item should not be null
                next(docs.map((doc) => this.fromFirestore(doc)));
            },
            error: (e) => error?.(e),
            complete: () => {
                complete?.();
            },
        });
    }
    async aggregate(aggregate) {
        const aggregateSpec = {};
        for (const [k, v] of Object.entries(aggregate.spec)) {
            switch (v.kind) {
                case 'count':
                    aggregateSpec[k] = count();
                    break;
                case 'sum':
                    aggregateSpec[k] = sum(v.path);
                    break;
                case 'average':
                    aggregateSpec[k] = average(v.path);
                    break;
                default:
                    return assertNever(v);
            }
        }
        const res = await getAggregateFromServer(toFirestoreQuery(this.db, aggregate.query), aggregateSpec);
        return res.data();
    }
    async set(doc, options) {
        const data = this.toFirestoreData(doc);
        await (options?.tx
            ? options.tx instanceof Transaction
                ? options.tx.set(this.docRef(doc), data)
                : options.tx.set(this.docRef(doc), data)
            : setDoc(this.docRef(doc), data));
    }
    async delete(id, options) {
        await (options?.tx ? options.tx.delete(this.docRef(id)) : deleteDoc(this.docRef(id)));
    }
    async batchSet(docs, options) {
        await this.batchWriteOperation(docs, {
            batch: (batch, doc) => batch.set(this.docRef(doc), this.toFirestoreData(doc)),
            transaction: (tx, doc) => tx.set(this.docRef(doc), this.toFirestoreData(doc)),
        }, options);
    }
    async batchDelete(ids, options) {
        await this.batchWriteOperation(ids, {
            batch: (batch, id) => batch.delete(this.docRef(id)),
            transaction: (tx, id) => tx.delete(this.docRef(id)),
        }, options);
    }
    async batchWriteOperation(targets, runner, options) {
        const tx = options?.tx;
        if (tx) {
            if (tx instanceof Transaction) {
                targets.forEach((target) => runner.transaction(tx, target));
            }
            else {
                targets.forEach((target) => runner.batch(tx, target));
            }
        }
        else {
            const batch = writeBatch(this.db);
            targets.forEach((target) => runner.batch(batch, target));
            await batch.commit();
        }
    }
    docRef(id) {
        return doc(this.db, docPath(this.collection, id));
    }
    collectionRef(parentId) {
        return collection(this.db, collectionPath(this.collection, parentId));
    }
    fromFirestore(doc) {
        const data = doc.data();
        const [id, ...parentPath] = docPathElements(doc.ref);
        return data
            ? {
                ...this.collection.data.from(data),
                ...this.collection.collectionPath.from(parentPath),
                ...this.collection.id.from(id.id),
            }
            : undefined;
    }
    toFirestoreData(data) {
        return this.collection.data.to(data);
    }
}
/**
 * Obtain document path elements from DocumentReference
 */
export const docPathElements = (doc) => {
    const parentPath = [];
    let cursor = doc.parent.parent;
    while (cursor) {
        parentPath.push({ id: cursor.id, collection: cursor.parent.id });
        cursor = cursor.parent.parent;
    }
    return [{ collection: doc.parent.id, id: doc.id }, ...parentPath];
};
export const toFirestoreQuery = (db, query) => {
    const { filter, nonFilter } = (query.constraints ?? []).reduce((acc, constraint) => {
        switch (constraint.kind) {
            case 'where': {
                const filter = toFirestoreQueryFilterConstraint(constraint.filter);
                acc.filter = acc.filter ? and(acc.filter, filter) : filter;
                break;
            }
            case 'orderBy':
                acc.nonFilter.push(orderBy(constraint.field, constraint.direction));
                break;
            case 'limit':
                acc.nonFilter.push(limit(constraint.limit));
                break;
            case 'limitToLast':
                acc.nonFilter.push(limitToLast(constraint.limit));
                break;
            case 'offset':
                // https://github.com/firebase/firebase-js-sdk/issues/479
                throw new Error('firestore-js-sdk does not support offset constraint');
            case 'startAt': {
                const { cursor } = constraint;
                acc.nonFilter.push(startAt(...cursor));
                break;
            }
            case 'startAfter': {
                const { cursor } = constraint;
                acc.nonFilter.push(startAfter(...cursor));
                break;
            }
            case 'endAt': {
                const { cursor } = constraint;
                acc.nonFilter.push(endAt(...cursor));
                break;
            }
            case 'endBefore': {
                const { cursor } = constraint;
                acc.nonFilter.push(endBefore(...cursor));
                break;
            }
            default:
                return assertNever(constraint);
        }
        return acc;
    }, { nonFilter: [] });
    let base;
    switch (query.base.kind) {
        case 'collection':
            base = collection(db, collectionPath(query.base.collection, query.base.parentId));
            break;
        case 'collectionGroup':
            base = collectionGroup(db, query.base.collection.name);
            break;
        case 'extends':
            base = toFirestoreQuery(db, query.base.query);
            break;
        default:
            base = assertNever(query.base);
    }
    return filter
        ? filter instanceof QueryCompositeFilterConstraint
            ? firestoreQuery(base, filter, ...nonFilter)
            : firestoreQuery(base, filter, ...nonFilter)
        : firestoreQuery(base, ...nonFilter);
};
const toFirestoreQueryFilterConstraint = (expr) => {
    switch (expr.kind) {
        case 'where':
            return where(expr.fieldPath, expr.opStr, expr.value);
        case 'or':
            return or(...expr.filters.map(toFirestoreQueryFilterConstraint));
        case 'and':
            return and(...expr.filters.map(toFirestoreQueryFilterConstraint));
        default:
            return assertNever(expr);
    }
};
